from django.apps import AppConfig

from service.decorators import SERVICE_APP_NAME, SERVICE_APP_DESCRIBE, SERVICE_API_IP, SERVICE_DEFAULT_SERVICE, \
    SERVICE_DEFAULT_METHODS, SERVICE_APP_ID


class ServiceConfig(AppConfig):
    name = 'service'

    def ready(self):
        """
        在子类中重写此方法，以便在Django启动时运行代码。
        """
        # 启动时检查环境变量是否设置
        if SERVICE_APP_ID is None:
            raise ValueError('Settings SERVICE_APP_ID unconfigured.')

        if SERVICE_APP_NAME is None:
            raise ValueError('Settings SERVICE_APP_NAME unconfigured.')

        if SERVICE_APP_DESCRIBE is None:
            raise ValueError('Settings SERVICE_APP_DESCRIBE unconfigured.')

        if SERVICE_API_IP is None:
            raise ValueError('Settings SERVICE_API_IP unconfigured.')

        if SERVICE_DEFAULT_SERVICE is None:
            raise ValueError('Settings SERVICE_DEFAULT_SERVICE unconfigured.')

        if SERVICE_DEFAULT_METHODS is None:
            raise ValueError('Settings SERVICE_DEFAULT_METHODS unconfigured.')



import logging

import requests

from approve.models import Permission
from permission.settings import KONG_SERVICES_PATH, KONG_ROUTES_PATH, KONG_PERMISSION_OPR_MAP, KONG_AUTH_TAG, \
    KONG_PERMISSION_MAP, KONG_ROUTE_PATH, KONG_SERVICE_PATH, KONG_SERVICE_JWT_PATH, KONG_CONSUMERS_PATH, \
    KONG_ROUTE_ACL_PATH, KONG_SUPER_GROUP

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)

logger.addHandler(ch)


def install_services(services):
    """
    安装服务
    """
    for service in services:
        try:
            result = requests.post(KONG_SERVICES_PATH, data=service, timeout=8)
            if result.status_code == 201:
                logger.info('Service {} {}'.format(service.get('name'), 'create success.'))
                if KONG_AUTH_TAG in service.get('tags', []):
                    data = {
                        'name': 'jwt',
                        'config.claims_to_verify': 'exp'
                    }

                    result = requests.post(KONG_SERVICE_JWT_PATH.format(service.get('name')), data=data, timeout=8)
                    if result.status_code == 201:
                        logger.info('Service {} JWT {}'.format(service.get('name'), 'create success.'))
                    else:
                        logger.info('Service {} JWT {}'.format(service.get('name'), 'create failed.'))
            else:
                logger.warning('Service {} {}'.format(service.get('name'), 'create failed.'))
        except requests.exceptions.RequestException:
            logger.error('API gateway request failed')


def install_routes(routes):
    """
    安装路由
    """
    for route in routes:
        for method in route.get('methods', []):
            opr = KONG_PERMISSION_OPR_MAP.get(method)
            if opr is not None:
                data = route.copy()
                del data['hans']
                data['name'] = '{}_{}_{}'.format(opr, data.get('service'), data.get('name'))
                data['methods'] = [method, 'OPTIONS']
                data['protocols'] = ['http']
                data['strip_path'] = 'false'

                try:
                    result = requests.post(KONG_ROUTES_PATH.format(data.get('service')), data=data, timeout=8)

                    if result.status_code == 201:
                        if KONG_AUTH_TAG in data.get('tags', []):
                            action = KONG_PERMISSION_MAP.get(method)
                            p = Permission(id=data.get('name'), name='{}{}'.format(action, route.get('hans')))
                            p.save()

                        logger.info('Route {} {}'.format(data.get('name'), 'create success.'))

                        if KONG_AUTH_TAG in data.get('tags', []):
                            sync_acl(data.get('name'), [KONG_SUPER_GROUP])
                    else:
                        logger.warning('Route {} {}'.format(data.get('name'), 'create failed.'))
                except requests.exceptions.RequestException:
                    logger.error('API gateway request failed')


def uninstall_services(services):
    """
    卸载服务
    """
    for service in services:
        try:
            result = requests.delete(KONG_SERVICE_PATH.format(service.get('name')), timeout=8)
            if result.status_code == 204:
                logger.info('Service {} {}'.format(service.get('name'), 'delete success.'))
            else:
                logger.warning('Service {} {}'.format(service.get('name'), 'delete failed.'))
        except requests.exceptions.RequestException:
            logger.error('API gateway request failed')


def uninstall_routes(routes):
    """
    卸载路由
    """
    for route in routes:
        for method in route.get('methods', []):
            opr = KONG_PERMISSION_OPR_MAP.get(method)
            if opr is not None:
                name = '{}_{}_{}'.format(opr, route.get('service'), route.get('name'))

                try:
                    result = requests.delete(KONG_ROUTE_PATH.format(name), timeout=8)

                    if result.status_code == 204:
                        p = Permission.objects.filter(id=name)
                        p.delete()
                        logger.info('Route {} {}'.format(name, 'delete success.'))
                    else:
                        logger.warning('Route {} {}'.format(name, 'delete failed.'))
                except requests.exceptions.RequestException:
                    logger.error('API gateway request failed')


def create_user(username):
    """
    创建用户
    """
    try:
        result = requests.post(KONG_CONSUMERS_PATH, data='')
        if result.status_code == 201:
            logger.info('Consumers {} {}'.format(username, 'create success.'))
        else:
            logger.warning('Consumers {} {}'.format(username, 'create failed.'))
    except requests.exceptions.RequestException:
        logger.error('API gateway request failed')


def sync_acl(route, group):
    """
    同步用户组
    """
    try:
        data = {
            'name': 'acl',
            'config.whitelist': ','.join(group)
        }
        result = requests.post(KONG_ROUTE_ACL_PATH.format(route), data=data, timeout=8)
        if result.status_code == 201:
            logger.info('Route {} acls {} {}'.format(route, ','.join(group), 'create success.'))
        else:
            logger.warning('Route {} acls {} {}'.format(route, ','.join(group), 'create failed.'))
    except requests.exceptions.RequestException:
        logger.error('API gateway request failed')



from django.conf import settings

METHODS_LIST = ['GET', 'POST', 'PUT', 'DELETE']
TAGS_LIST = ['authentication', 'anonymous']

SERVICE_APP_ID = getattr(settings, 'SERVICE_APP_ID', None)
SERVICE_APP_NAME = getattr(settings, 'SERVICE_APP_NAME', None)
SERVICE_APP_DESCRIBE = getattr(settings, 'SERVICE_APP_DESCRIBE', None)

SERVICE_API_PROTOCOL = getattr(settings, 'SERVICE_API_PROTOCOL', ['http'])
SERVICE_API_IP = getattr(settings, 'SERVICE_API_IP', None)
SERVICE_API_PORT = getattr(settings, 'SERVICE_API_PORT', 80)

SERVICE_DEFAULT_SERVICE = getattr(settings, 'SERVICE_DEFAULT_SERVICE', None)
SERVICE_DEFAULT_METHODS = getattr(settings, 'SERVICE_DEFAULT_METHODS', ['GET', 'POST', 'PUT', 'DELETE'])
SERVICE_DEFAULT_TAGS = getattr(settings, 'SERVICE_DEFAULT_TAGS', ['authentication'])

services = [
    {
        'name': SERVICE_DEFAULT_SERVICE,
        'protocol': SERVICE_API_PROTOCOL,
        'host': SERVICE_API_IP,
        'port': SERVICE_API_PORT,
        'path': '/',
        'tags': SERVICE_DEFAULT_TAGS
    }
]

routes = []


def register(name, paths, methods=None, service=None, tags=None, hans=None):
    """
    """
    def wrapper(clazz):
        if isinstance(name, str) is not True:
            raise ValueError('The name is not a string.')

        if isinstance(paths, list) is not True:
            raise ValueError('The paths is not a list.')

        if methods is not None and set(methods).issubset(set(METHODS_LIST)) is not True:
            raise ValueError('The methods is not legal.')

        if tags is not None and set(tags).issubset(set(TAGS_LIST)) is not True:
            raise ValueError('The tags is not legal.')

        if service is not None and isinstance(service, str) is not True:
            raise ValueError('The service is not a string.')

        if hans is not None and isinstance(hans, str) is not True:
            raise ValueError('The hans is not a string.')

        if service is not None:
            if service not in [s.get('name') for s in services]:
                s = {
                    'name': service,
                    'protocol': SERVICE_API_PROTOCOL,
                    'host': SERVICE_API_IP,
                    'port': SERVICE_API_PORT,
                    'path': '/',
                    'tags': ['authentication']
                }

                if tags is not None:
                    s['tags'] = tags

                services.append(s)

        route = {
            'name': name,
            'methods': methods or SERVICE_DEFAULT_METHODS,
            'tags': tags or SERVICE_DEFAULT_TAGS,
            'paths': paths,
            'service': service or SERVICE_DEFAULT_SERVICE,
            'hans': hans or name
        }
        routes.append(route)

        return clazz
    return wrapper



from django.http import JsonResponse
from service.decorators import register, SERVICE_APP_NAME, SERVICE_APP_DESCRIBE, services, routes, SERVICE_APP_ID


@register('endpoint', ['/api/endpoint'], methods=['GET'], tags=['anonymous'], hans='服务信息')
def endpoint(request):
    """
    服务配置
    """
    data = {
        'id': SERVICE_APP_ID,
        'name': SERVICE_APP_NAME,
        'describe': SERVICE_APP_DESCRIBE,
        'config': {'services': services, 'routes': routes}
    }
    return JsonResponse(data=data)
